@@ -19,10 +19,42 @@
 #include "main.h"
 #include "prefdb.h"
 #include <glibmm/i18n.h>
-#include "grabber.h" // TODO???
+#include "grabber.h"
 
 #include <typeinfo>
 
+bool TreeViewMulti::on_button_press_event(GdkEventButton* event) {
+	int cell_x, cell_y;
+	Gtk::TreeViewColumn *column;
+	pending = (get_path_at_pos(event->x, event->y, path, column, cell_x, cell_y))
+		&& (get_selection()->is_selected(path))
+		&& !(event->state & (GDK_CONTROL_MASK|GDK_SHIFT_MASK));
+	return Gtk::TreeView::on_button_press_event(event);
+}
+
+bool TreeViewMulti::on_button_release_event(GdkEventButton* event) {
+	if (pending) {
+		pending = false;
+		get_selection()->unselect_all();
+		get_selection()->select(path);
+	}
+	return Gtk::TreeView::on_button_release_event(event);
+}
+
+void TreeViewMulti::on_drag_begin(const Glib::RefPtr<Gdk::DragContext> &context) {
+	pending = false;
+	if (get_selection()->count_selected_rows() <= 1)
+		return Gtk::TreeView::on_drag_begin(context);
+	Glib::RefPtr<Gdk::Pixbuf> pb = render_icon(Gtk::Stock::DND_MULTIPLE, Gtk::ICON_SIZE_DND);
+	context->set_icon(pb, pb->get_width(), pb->get_height());
+}
+
+bool negate(bool b) { return !b; }
+
+TreeViewMulti::TreeViewMulti() : Gtk::TreeView(), pending(false) {
+	get_selection()->set_select_function(sigc::group(&negate, sigc::ref(pending)));
+}
+
 class CellEditableAccel : public Gtk::EventBox, public Gtk::CellEditable {
 	CellRendererTextish *parent;
 	Glib::ustring path;
@@ -151,14 +183,16 @@ const char *type_info_to_name(const std::type_info *info) {
 }
 
 Actions::Actions() :
-	tv(0),
 	apps_view(0),
 	editing_new(false),
 	editing(false),
 	action_list(actions.get_root())
 {
-	widgets->get_widget("treeview_actions", tv);
+	Gtk::ScrolledWindow *sw;
+	widgets->get_widget("scrolledwindow_actions", sw);
 	widgets->get_widget("treeview_apps", apps_view);
+	sw->add(tv);
+	tv.show();
 
 	Gtk::Button *button_add, *button_add_app, *button_add_group;
 	widgets->get_widget("button_add_action", button_add);
@@ -178,29 +212,29 @@ Actions::Actions() :
 	button_remove_app->signal_clicked().connect(sigc::mem_fun(*this, &Actions::on_remove_app));
 	button_reset_actions->signal_clicked().connect(sigc::mem_fun(*this, &Actions::on_reset_actions));
 
-	tv->signal_cursor_changed().connect(sigc::mem_fun(*this, &Actions::on_cursor_changed));
-	tv->signal_row_activated().connect(sigc::mem_fun(*this, &Actions::on_row_activated));
-	tv->get_selection()->signal_changed().connect(sigc::mem_fun(*this, &Actions::on_selection_changed));
+	tv.signal_cursor_changed().connect(sigc::mem_fun(*this, &Actions::on_cursor_changed));
+	tv.signal_row_activated().connect(sigc::mem_fun(*this, &Actions::on_row_activated));
+	tv.get_selection()->signal_changed().connect(sigc::mem_fun(*this, &Actions::on_selection_changed));
 
 	tm = Store::create(cols, this);
 
-	tv->get_selection()->set_mode(Gtk::SELECTION_MULTIPLE);
-	tv->set_row_separator_func(sigc::mem_fun(*this, &Actions::on_row_separator));
+	tv.get_selection()->set_mode(Gtk::SELECTION_MULTIPLE);
+	tv.set_row_separator_func(sigc::mem_fun(*this, &Actions::on_row_separator));
 
 	int n;
-	n = tv->append_column(_("Stroke"), cols.stroke);
-	tv->get_column(n-1)->set_sort_column(cols.id);
+	n = tv.append_column(_("Stroke"), cols.stroke);
+	tv.get_column(n-1)->set_sort_column(cols.id);
 	tm->set_sort_func(cols.id, sigc::mem_fun(*this, &Actions::compare_ids));
 	tm->set_default_sort_func(sigc::mem_fun(*this, &Actions::compare_ids));
 	tm->set_sort_column(Gtk::TreeSortable::DEFAULT_SORT_COLUMN_ID, Gtk::SORT_ASCENDING);
 
-	n = tv->append_column(_("Name"), cols.name);
-	Gtk::CellRendererText *name_renderer = dynamic_cast<Gtk::CellRendererText *>(tv->get_column_cell_renderer(n-1));
+	n = tv.append_column(_("Name"), cols.name);
+	Gtk::CellRendererText *name_renderer = dynamic_cast<Gtk::CellRendererText *>(tv.get_column_cell_renderer(n-1));
 	name_renderer->property_editable() = true;
 	name_renderer->signal_edited().connect(sigc::mem_fun(*this, &Actions::on_name_edited));
 	name_renderer->signal_editing_started().connect(sigc::mem_fun(*this, &Actions::on_something_editing_started));
 	name_renderer->signal_editing_canceled().connect(sigc::mem_fun(*this, &Actions::on_something_editing_canceled));
-	Gtk::TreeView::Column *col_name = tv->get_column(n-1);
+	Gtk::TreeView::Column *col_name = tv.get_column(n-1);
 	col_name->set_sort_column(cols.name);
 //	col_name->add_attribute(name_renderer->property_text(), cols.name);
 	col_name->set_cell_data_func(*name_renderer, sigc::mem_fun(*this, &Actions::on_cell_data_name));
@@ -218,14 +252,14 @@ Actions::Actions() :
 	type_renderer->signal_editing_started().connect(sigc::mem_fun(*this, &Actions::on_something_editing_started));
 	type_renderer->signal_editing_canceled().connect(sigc::mem_fun(*this, &Actions::on_something_editing_canceled));
 
-	n = tv->append_column(_("Type"), *type_renderer);
-	Gtk::TreeView::Column *col_type = tv->get_column(n-1);
+	n = tv.append_column(_("Type"), *type_renderer);
+	Gtk::TreeView::Column *col_type = tv.get_column(n-1);
 	col_type->add_attribute(type_renderer->property_text(), cols.type);
 	col_type->set_cell_data_func(*type_renderer, sigc::mem_fun(*this, &Actions::on_cell_data_type));
 
 	CellRendererTextish *arg_renderer = Gtk::manage(new CellRendererTextish);
-	n = tv->append_column(_("Details"), *arg_renderer);
-	Gtk::TreeView::Column *col_arg = tv->get_column(n-1);
+	n = tv.append_column(_("Details"), *arg_renderer);
+	Gtk::TreeView::Column *col_arg = tv.get_column(n-1);
 	col_arg->add_attribute(arg_renderer->property_text(), cols.arg);
 	col_arg->set_cell_data_func(*arg_renderer, sigc::mem_fun(*this, &Actions::on_cell_data_arg));
 	arg_renderer->property_editable() = true;
@@ -236,9 +270,9 @@ Actions::Actions() :
 	arg_renderer->items = Misc::types;
 
 	update_action_list();
-	tv->set_model(tm);
-	tv->enable_model_drag_source();
-	tv->enable_model_drag_dest();
+	tv.set_model(tm);
+	tv.enable_model_drag_source();
+	tv.enable_model_drag_dest();
 
 	check_show_deleted->signal_toggled().connect(sigc::mem_fun(*this, &Actions::update_action_list));
 	expander_apps->property_expanded().signal_changed().connect(sigc::mem_fun(*this, &Actions::on_apps_selection_changed));
@@ -357,9 +391,20 @@ bool Actions::AppsStore::drag_data_received_vfunc(const Gtk::TreeModel::Path &de
 	ActionListDiff *actions = dest_iter ? (*dest_iter)[parent->ca.actions] : (ActionListDiff *)NULL;
 	if (!actions || actions == parent->action_list)
 		return false;
-	RStrokeInfo si = parent->action_list->get_info(src_id);
-	parent->action_list->remove(src_id);
-	actions->add(*si);
+	Glib::RefPtr<Gtk::TreeSelection> sel = parent->tv.get_selection();
+	if (sel->count_selected_rows() <= 1) {
+		RStrokeInfo si = parent->action_list->get_info(src_id);
+		parent->action_list->remove(src_id);
+		actions->add(*si);
+	} else {
+		std::vector<Gtk::TreePath> paths = sel->get_selected_rows();
+		for (std::vector<Gtk::TreePath>::iterator i = paths.begin(); i != paths.end(); ++i) {
+			Unique *id = (*parent->tm->get_iter(*i))[parent->cols.id];
+			RStrokeInfo si = parent->action_list->get_info(id);
+			parent->action_list->remove(id);
+			actions->add(*si);
+		}
+	}
 	parent->update_action_list();
 	update_actions();
 	return true;
@@ -373,8 +418,19 @@ bool Actions::Store::row_draggable_vfunc(const Gtk::TreeModel::Path &path) const
 		return false;
 	if (sort != Gtk::SORT_ASCENDING)
 		return false;
-	Unique *id = (*parent->tm->get_iter(path))[parent->cols.id];
-	return id->level == parent->action_list->level;
+	Glib::RefPtr<Gtk::TreeSelection> sel = parent->tv.get_selection();
+	if (sel->count_selected_rows() <= 1) {
+		Unique *id = (*parent->tm->get_iter(path))[parent->cols.id];
+		return id->level == parent->action_list->level;
+	} else {
+		std::vector<Gtk::TreePath> paths = sel->get_selected_rows();
+		for (std::vector<Gtk::TreePath>::iterator i = paths.begin(); i != paths.end(); ++i) {
+			Unique *id = (*parent->tm->get_iter(*i))[parent->cols.id];
+			if (id->level != parent->action_list->level)
+				return false;
+		}
+		return true;
+	}
 }
 
 bool Actions::Store::row_drop_possible_vfunc(const Gtk::TreeModel::Path &dest, const Gtk::SelectionData &selection) const {
@@ -413,15 +469,30 @@ bool Actions::Store::drag_data_received_vfunc(const Gtk::TreeModel::Path &dest,
 	Unique *dest_id = dest_iter ? (*dest_iter)[parent->cols.id] : (Unique *)0;
 	if (dest_id && src_id->level != dest_id->level)
 		return false;
-	if (parent->action_list->move(src_id, dest_id)) {
-		(*parent->tm->get_iter(src))[parent->cols.id] = src_id;
-		update_actions();
+	Glib::RefPtr<Gtk::TreeSelection> sel = parent->tv.get_selection();
+	if (sel->count_selected_rows() <= 1) {
+		if (parent->action_list->move(src_id, dest_id)) {
+			(*parent->tm->get_iter(src))[parent->cols.id] = src_id;
+			update_actions();
+		}
+	} else {
+		std::vector<Gtk::TreePath> paths = sel->get_selected_rows();
+		bool updated = false;
+		for (std::vector<Gtk::TreePath>::iterator i = paths.begin(); i != paths.end(); ++i) {
+			Unique *id = (*parent->tm->get_iter(*i))[parent->cols.id];
+			if (parent->action_list->move(id, dest_id))
+				updated = true;
+		}
+		if (updated) {
+			parent->update_action_list();
+			update_actions();
+		}
 	}
 	return false;
 }
 
 void Actions::on_type_edited(const Glib::ustring &path, const Glib::ustring &new_text) {
-	tv->grab_focus();
+	tv.grab_focus();
 	Gtk::TreeRow row(*tm->get_iter(path));
 	Type new_type = from_name(new_text);
 	Type old_type = from_name(row[cols.type]);
@@ -473,7 +544,7 @@ void Actions::on_type_edited(const Glib::ustring &path, const Glib::ustring &new
 }
 
 void Actions::on_button_delete() {
-	int n = tv->get_selection()->count_selected_rows();
+	int n = tv.get_selection()->count_selected_rows();
 
 	Glib::ustring str;
 	if (n == 1)
@@ -496,7 +567,7 @@ void Actions::on_button_delete() {
 	if (!ok)
 		return;
 
-	std::vector<Gtk::TreePath> paths = tv->get_selection()->get_selected_rows();
+	std::vector<Gtk::TreePath> paths = tv.get_selection()->get_selected_rows();
 	for (std::vector<Gtk::TreePath>::iterator i = paths.begin(); i != paths.end(); ++i) {
 		Gtk::TreeRow row(*tm->get_iter(*i));
 		action_list->remove(row[cols.id]);
@@ -572,7 +643,7 @@ void Actions::on_remove_app() {
 }
 
 void Actions::on_reset_actions() {
-	std::vector<Gtk::TreePath> paths = tv->get_selection()->get_selected_rows();
+	std::vector<Gtk::TreePath> paths = tv.get_selection()->get_selected_rows();
 	for (std::vector<Gtk::TreePath>::iterator i = paths.begin(); i != paths.end(); ++i) {
 		Gtk::TreeRow row(*tm->get_iter(*i));
 		action_list->reset(row[cols.id]);
@@ -664,6 +735,7 @@ void Actions::update_row(const Gtk::TreeRow &row) {
 }
 
 extern boost::shared_ptr<sigc::slot<void, RStroke> > stroke_action;
+Source<ActionListDiff *> stroke_app(NULL);
 
 class Actions::OnStroke {
 	Actions *parent;
@@ -699,6 +771,7 @@ public:
 		stroke = stroke_;
 		Glib::signal_idle().connect(sigc::mem_fun(*this, &OnStroke::run));
 		stroke_action.reset();
+		stroke_app.set(NULL);
 	}
 };
 
@@ -726,12 +799,14 @@ void Actions::on_row_activated(const Gtk::TreeModel::Path& path, Gtk::TreeViewCo
 
 	OnStroke ps(this, dialog, row);
 	stroke_action.reset(new sigc::slot<void, RStroke>(sigc::mem_fun(ps, &OnStroke::delayed_run)));
+	stroke_app.set(action_list);
 
 	dialog->show();
 	cancel->grab_focus();
 	int response = dialog->run();
 	dialog->hide();
 	stroke_action.reset();
+	stroke_app.set(NULL);
 	if (response != 1)
 		return;
 
@@ -744,29 +819,29 @@ void Actions::on_row_activated(const Gtk::TreeModel::Path& path, Gtk::TreeViewCo
 void Actions::on_button_record() {
 	Gtk::TreeModel::Path path;
 	Gtk::TreeViewColumn *col;
-	tv->get_cursor(path, col);
+	tv.get_cursor(path, col);
 	on_row_activated(path, col);
 }
 
 void Actions::on_cursor_changed() {
 	Gtk::TreeModel::Path path;
 	Gtk::TreeViewColumn *col;
-	tv->get_cursor(path, col);
+	tv.get_cursor(path, col);
 	Gtk::TreeRow row(*tm->get_iter(path));
 }
 
 Gtk::TreeRow Actions::get_selected_row() {
-	std::vector<Gtk::TreePath> paths = tv->get_selection()->get_selected_rows();
+	std::vector<Gtk::TreePath> paths = tv.get_selection()->get_selected_rows();
 	return Gtk::TreeRow(*tm->get_iter(*paths.begin()));
 }
 
 void Actions::on_selection_changed() {
-	int n = tv->get_selection()->count_selected_rows();
+	int n = tv.get_selection()->count_selected_rows();
 	button_record->set_sensitive(n == 1);
 	button_delete->set_sensitive(n >= 1);
 	bool resettable = false;
 	if (n) {
-		std::vector<Gtk::TreePath> paths = tv->get_selection()->get_selected_rows();
+		std::vector<Gtk::TreePath> paths = tv.get_selection()->get_selected_rows();
 		for (std::vector<Gtk::TreePath>::iterator i = paths.begin(); i != paths.end(); ++i) {
 			Gtk::TreeRow row(*tm->get_iter(*i));
 			if (action_list->resettable(row[cols.id])) {
@@ -781,8 +856,8 @@ void Actions::on_selection_changed() {
 void Actions::on_button_new() {
 	editing_new = true;
 	Unique *before = 0;
-	if (tv->get_selection()->count_selected_rows()) {
-		std::vector<Gtk::TreePath> paths = tv->get_selection()->get_selected_rows();
+	if (tv.get_selection()->count_selected_rows()) {
+		std::vector<Gtk::TreePath> paths = tv.get_selection()->get_selected_rows();
 		Gtk::TreeIter i = tm->get_iter(paths[paths.size()-1]);
 		i++;
 		if (i != tm->children().end())
@@ -810,7 +885,7 @@ bool Actions::do_focus(Unique *id, Gtk::TreeViewColumn *col, bool edit) {
 		Gtk::TreeModel::Children chs = tm->children();
 		for (Gtk::TreeIter i = chs.begin(); i != chs.end(); ++i)
 			if ((*i)[cols.id] == id) {
-				tv->set_cursor(Gtk::TreePath(*i), *col, edit);
+				tv.set_cursor(Gtk::TreePath(*i), *col, edit);
 			}
 	}
 	return false;
@@ -818,7 +893,7 @@ bool Actions::do_focus(Unique *id, Gtk::TreeViewColumn *col, bool edit) {
 
 void Actions::focus(Unique *id, int col, bool edit) {
 	editing = false;
-	Glib::signal_idle().connect(sigc::bind(sigc::mem_fun(*this, &Actions::do_focus), id, tv->get_column(col), edit));
+	Glib::signal_idle().connect(sigc::bind(sigc::mem_fun(*this, &Actions::do_focus), id, tv.get_column(col), edit));
 }
 
 void Actions::on_name_edited(const Glib::ustring& path, const Glib::ustring& new_text) {
@@ -889,7 +964,7 @@ void Actions::on_something_editing_started(Gtk::CellEditable* editable, const Gl
 }
 
 void Actions::on_arg_editing_started(Gtk::CellEditable* editable, const Glib::ustring& path) {
-	tv->grab_focus();
+	tv.grab_focus();
 	Gtk::TreeRow row(*tm->get_iter(path));
 	if (from_name(row[cols.type]) != BUTTON)
 		return;
