@@ -45,7 +45,7 @@ public:
 class Grabber;
 extern Grabber *grabber;
 
-float rescaleValuatorAxis(int coord, int fmin, int fmax, int tmax);
+float rescaleValuatorAxis(int coord, int fmin, int fmax, int tmin, int tmax, int defmax);
 bool has_wm_state(Window w);
 bool has_atom(Window w, Atom prop, Atom value);
 
@@ -69,6 +69,13 @@ public:
 	void select_proximity();
 
 	struct XiDevice {
+		class OpenException : public std::exception {
+			char *msg;
+			public:
+			OpenException(const char *name);
+			virtual const char* what() const throw() { return msg ? msg : "Open Failed"; }
+			~OpenException() throw() { free(msg); }
+		};
 		std::string name;
 		std::map<guint, guint> inv_map;
 		XDevice *dev;
@@ -77,10 +84,12 @@ public:
 		bool supports_proximity, supports_pressure;
 		bool active;
 		int pressure_min, pressure_max;
-		int min_x, max_x, min_y, max_y;
+		int min[2], max[2];
 		bool absolute;
 		int valuators[2];
 		int num_buttons;
+		XiDevice(Grabber *, XDeviceInfo *);
+		~XiDevice();
 		int normalize_pressure(int pressure) {
 			return 255 * (pressure - pressure_min) / (pressure_max - pressure_min);
 		}
@@ -89,7 +98,8 @@ public:
 		void grab_device(bool grab);
 		void grab_button(ButtonInfo &bi, bool grab);
 		void update_pointer_mapping();
-		void translate_coords(int *axis_data, float &x, float &y);
+		void update_axes();
+		void translate_coords(int sx, int sy, int *axis_data, float &x, float &y);
 		bool translate_known_coords(int sx, int sy, int *axis_data, float &x, float &y);
 	};
 
@@ -99,8 +109,8 @@ public:
 	int event_presence;
 	XEventClass presence_class;
 
-	XiDevice **xi_devs;
-	int xi_devs_n;
+	typedef std::map<XID, boost::shared_ptr<XiDevice> > DeviceMap;
+	DeviceMap xi_devs;
 	int nMajor;
 private:
 	bool init_xi();
@@ -120,8 +130,6 @@ private:
 	void grab_xi(bool);
 	void regrab_xi();
 	void grab_xi_devs(bool);
-	std::string get_wm_class(Window w);
-	std::string wm_class;
 
 	void update_excluded();
 
@@ -133,7 +141,7 @@ public:
 	Grabber();
 	~Grabber();
 	bool handle(XEvent &ev) { return children.handle(ev); }
-	std::string get_wm_class() { return wm_class; }
+	Out<std::string> *current_class;
 
 	void queue_suspend() { queue(sigc::mem_fun(*this, &Grabber::suspend)); }
 	void queue_resume() { queue(sigc::mem_fun(*this, &Grabber::resume)); }
