@@ -58,7 +58,6 @@ Grabber::XiDevice *current_dev = 0;
 std::set<guint> xinput_pressed;
 
 static bool show_gui = false;
-static bool rotated = false;
 static int offset_x = 0;
 static int offset_y = 0;
 
@@ -274,7 +273,6 @@ static int xErrorHandler(Display *dpy2, XErrorEvent *e) {
 			 e->minor_code == X_GrabDeviceButton)) {
 		if (!handler || Handler::idle()) {
 			printf(_("Error: %s\n"), e->request_code==X_GrabButton ? "A grab failed" : "An XInput grab failed");
-			printf(_("Is easystroke already running?\n"));
 		} else {
 			printf(_("Error: A grab failed.  Resetting...\n"));
 			bail_out();
@@ -753,7 +751,7 @@ public:
 
 		std::map<guint, RAction> as;
 		std::map<guint, RRanking> rs;
-		actions.get_action_list(grabber->get_wm_class())->handle_advanced(s, as, rs, b1, b2);
+		actions.get_action_list(grabber->current_class->get())->handle_advanced(s, as, rs, b1, b2);
 		if (press_t) {
 			if (as.count(b2))
 				discard(press_t);
@@ -1124,7 +1122,7 @@ protected:
 			return parent->replace_child(NULL);
 		}
 		RRanking ranking;
-		RAction act = actions.get_action_list(grabber->get_wm_class())->handle(s, ranking);
+		RAction act = actions.get_action_list(grabber->current_class->get())->handle(s, ranking);
 		if (!IS_CLICK(act))
 			Ranking::queue_show(ranking, e);
 		if (!act) {
@@ -1265,6 +1263,10 @@ static void do_run_by_name(RAction act) {
 }
 
 void run_by_name(const char *str) {
+	if (!strcmp(str, "")) {
+		win->show_hide();
+		return;
+	}
 	for (ActionDB::const_iterator i = actions.begin(); i != actions.end(); i++) {
 		if (i->second.name == std::string(str)) {
 			if (i->second.action)
@@ -1275,10 +1277,6 @@ void run_by_name(const char *str) {
 	printf(_("Warning: No action \"%s\" defined\n"), str);
 }
 
-bool disable_root() {
-	return !actions.get_root()->order_size() && actions.get_root()->size_rec();
-}
-
 void icon_warning() {
 	for (ActionDB::const_iterator i = actions.begin(); i != actions.end(); i++) {
 		Misc *m = dynamic_cast<Misc *>(i->second.action.get());
@@ -1347,7 +1345,7 @@ static void link_button_hook(Gtk::LinkButton *, const Glib::ustring& uri) { xdg_
 static void about_dialog_hook(Gtk::AboutDialog &, const Glib::ustring& url) { xdg_open(url); }
 
 // dbus-send --type=method_call --dest=org.easystroke /org/easystroke org.easystroke.send string:"foo"
-static void send_dbus(char *str) {
+static void send_dbus(const char *str) {
 	GError *error = 0;
 	DBusGConnection *bus = dbus_g_bus_get(DBUS_BUS_SESSION, &error);
 	if (!bus) {
@@ -1358,7 +1356,7 @@ static void send_dbus(char *str) {
 	dbus_g_proxy_call_no_reply(proxy, "send", G_TYPE_STRING, str, G_TYPE_INVALID);
 }
 
-bool start_dbus();
+int start_dbus();
 
 Main::Main() : kit(0) {
 	bindtextdomain("easystroke", is_dir("po") ? "po" : LOCALEDIR);
@@ -1392,6 +1390,11 @@ Main::Main() : kit(0) {
 		printf(_("Couldn't open display.\n"));
 		exit(EXIT_FAILURE);
 	}
+	if (start_dbus() < 0) {
+		printf(_("Easystroke is already running, showing configuration window instead.\n"));
+		send_dbus("");
+		exit(EXIT_SUCCESS);
+	}
 	ROOT = DefaultRootWindow(dpy);
 
 	ping_window = XCreateSimpleWindow(dpy, ROOT, 0, 0, 1, 1, 0, 0, 0);
@@ -1417,7 +1420,6 @@ Main::Main() : kit(0) {
 	handler->init();
 	XTestGrabControl(dpy, True);
 
-	start_dbus();
 }
 
 extern const char *gui_buffer;
@@ -1568,63 +1570,58 @@ void Main::create_config_dir() {
 
 extern Window get_app_window(Window &w);
 
-void Grabber::XiDevice::translate_coords(int *axis_data, float &x, float &y) {
-	if (!absolute) {
-		valuators[0] += axis_data[0];
-		valuators[1] += axis_data[1];
-		x = valuators[0];
-		y = valuators[1];
-		return;
-	}
-	valuators[0] = axis_data[0];
-	valuators[1] = axis_data[1];
-	int w = gdk_screen_width() - 1;
-	int h = gdk_screen_height() - 1;
-	if (!rotated) {
-		x = rescaleValuatorAxis(axis_data[0], min_x, max_x, w);
-		y = rescaleValuatorAxis(axis_data[1], min_y, max_y, h);
+void Grabber::XiDevice::translate_coords(int sx, int sy, int *axis_data, float &x, float &y) {
+	if (absolute || !xi_15) {
+		valuators[0] = axis_data[0];
+		valuators[1] = axis_data[1];
+		int w = gdk_screen_width() - 1;
+		int h = gdk_screen_height() - 1;
+		x = rescaleValuatorAxis(valuators[0], min[0], max[0], 0, w, w);
+		y = rescaleValuatorAxis(valuators[1], min[1], max[1], 0, h, h);
 	} else {
-		x = rescaleValuatorAxis(axis_data[0], min_y, max_y, w);
-		y = rescaleValuatorAxis(axis_data[1], min_x, max_x, h);
+		int w = gdk_screen_width() - 1;
+		int h = gdk_screen_height() - 1;
+		valuators[0] = rescaleValuatorAxis(sx, 0, w, min[0], max[0], w) + axis_data[0];
+		valuators[1] = rescaleValuatorAxis(sy, 0, h, min[1], max[1], h) + axis_data[1];
+		x = sx;
+		y = sy;
 	}
 }
 
 bool Grabber::XiDevice::translate_known_coords(int sx, int sy, int *axis_data, float &x, float &y) {
 	sx += offset_x;
 	sy += offset_y;
-	if (!absolute) {
-		valuators[0] = sx;
-		valuators[1] = sy;
-		x = valuators[0];
-		y = valuators[1];
-		return true;
-	}
-	valuators[0] = axis_data[0];
-	valuators[1] = axis_data[1];
 	int w = gdk_screen_width() - 1;
 	int h = gdk_screen_height() - 1;
-	x        = rescaleValuatorAxis(axis_data[0], min_x, max_x, w);
-	y        = rescaleValuatorAxis(axis_data[1], min_y, max_y, h);
-	if (axis_data[0] == sx && axis_data[1] == sy)
-		return true;
-	float x2 = rescaleValuatorAxis(axis_data[0], min_y, max_y, w);
-	float y2 = rescaleValuatorAxis(axis_data[1], min_x, max_x, h);
-	float d  = hypot(x - sx, y - sy);
-	float d2 = hypot(x2 - sx, y2 - sy);
-	if (d > 2 && d2 > 2) {
-		x = sx;
-		y = sy;
-		return false;
+
+	bool second_try = false;
+	while (true) {
+		if (absolute || !xi_15) {
+			valuators[0] = axis_data[0];
+			valuators[1] = axis_data[1];
+			x = rescaleValuatorAxis(axis_data[0], min[0], max[0], 0, w, w);
+			y = rescaleValuatorAxis(axis_data[1], min[1], max[1], 0, h, h);
+			if (hypot(x - sx, y - sy) < 50.0)
+				return true;
+		} else {
+			valuators[0] = rescaleValuatorAxis(sx, 0, w, min[0], max[0], w) + axis_data[0];
+			valuators[1] = rescaleValuatorAxis(sy, 0, h, min[1], max[1], h) + axis_data[1];
+			x = sx;
+			y = sy;
+			return true;
+		}
+		if (second_try) {
+			x = sx;
+			y = sy;
+			if (verbosity >= 1)
+				printf("Warning: Reloading Axis information failed\n");
+			return false;
+		}
+		if (verbosity >= 2)
+			printf("Reloading Axis information\n");
+		update_axes();
+		second_try = true;
 	}
-	if (d > 2)
-		rotated = true;
-	if (d2 > 2)
-		rotated = false;
-	if (rotated) {
-		x = x2;
-		y = y2;
-	}
-	return true;
 }
 
 void Main::handle_enter_leave(XEvent &ev) {
@@ -1776,8 +1773,8 @@ void Main::handle_event(XEvent &ev) {
 			return;
 		xinput_pressed.insert(bev->button);
 		float x, y;
-		if (xi_15 && xinput_pressed.size() > 1)
-			current_dev->translate_coords(bev->axis_data, x, y);
+		if (xinput_pressed.size() > 1)
+			current_dev->translate_coords(bev->x, bev->y, bev->axis_data, x, y);
 		else
 			current_dev->translate_known_coords(bev->x, bev->y, bev->axis_data, x, y);
 		H->press(bev->button, create_triple(x, y, bev->time));
@@ -1793,10 +1790,7 @@ void Main::handle_event(XEvent &ev) {
 			return;
 		xinput_pressed.erase(bev->button);
 		float x, y;
-		if (xi_15)
-			current_dev->translate_coords(bev->axis_data, x, y);
-		else
-			current_dev->translate_known_coords(bev->x, bev->y, bev->axis_data, x, y);
+		current_dev->translate_coords(bev->x, bev->y, bev->axis_data, x, y);
 
 		H->release(bev->button, create_triple(x, y, bev->time));
 		return;
@@ -1810,10 +1804,7 @@ void Main::handle_event(XEvent &ev) {
 		if (!current_dev || current_dev->dev->device_id != mev->deviceid)
 			return;
 		float x, y;
-		if (xi_15)
-			current_dev->translate_coords(mev->axis_data, x, y);
-		else
-			current_dev->translate_known_coords(mev->x, mev->y, mev->axis_data, x, y);
+		current_dev->translate_coords(mev->x, mev->y, mev->axis_data, x, y);
 		int z = 0;
 		if (current_dev->supports_pressure)
 			z = current_dev->normalize_pressure(mev->axis_data[2]);
@@ -1893,7 +1884,7 @@ std::string select_window() {
 	queue(sigc::ptr_fun(&SelectHandler::create));
 	gtk_main();
 	win->get_window().raise();
-	return grabber->get_wm_class();
+	return grabber->current_class->get();
 }
 
 void Button::run() {
