@@ -50,7 +50,7 @@ PrefDB::PrefDB() :
 	timeout_gestures(false),
 	tray_icon(true),
 	color(Gdk::Color("#980101")),
-	trace_width(5),
+	trace_width(3),
 	advanced_popups(true),
 	scroll_invert(true),
 	scroll_speed(2.0),
@@ -228,7 +228,9 @@ void PrefDB::init() {
 		exceptions.unsafe_ref().erase(i);
 		exceptions.unsafe_ref()[""] = bi;
 	}
-	new TimeoutProfile(prefs.timeout_profile);
+	new TimeoutProfile(timeout_profile);
+	watch(init_timeout);
+	watch(final_timeout);
 }
 
 PrefDB prefs;
